<?php

namespace Drupal\commerce_cib\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_cib\EncryptionInterface;
use Drupal\commerce_cib\Event\CibEvents;
use Drupal\commerce_cib\Event\FailedPayment;
use Drupal\commerce_cib\Event\NoCommunication;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannel;
use GuzzleHttp\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the CIBt payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "cib",
 *   label = "CIB",
 *   display_label = "CIB",
 *   payment_type = "payment_cib",
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_cib\PluginForm\CibForm",
 *   },
 * )
 */
class Cib extends OffsitePaymentGatewayBase implements CibInterface {

  // Default value for transaction timeouts in seconds.
  const TIMEOUT = 700;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The encryption service.
   *
   * @var \Drupal\commerce_cib\EncryptionInterface
   */
  protected $encryption;

  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannel
   */
  protected $logger;

  /**
   * The event dispatcher.
   *
   * @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher
   */
  protected $eventDispatcher;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time, ClientInterface $client, EncryptionInterface $encryption, LoggerChannel $logger, ContainerAwareEventDispatcher $event_dispatcher) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    $this->httpClient = $client;
    $this->encryption = $encryption;
    $path = $this->getMode() === 'live' ? $this->configuration['des-live'] : $this->configuration['des-test'];
    $this->encryption->setKeyfile($path);
    $this->logger = $logger;
    $this->eventDispatcher = $event_dispatcher;
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('http_client'),
      $container->get('commerce_cib.encryption'),
      $container->get('commerce_cib.logger'),
      $container->get('event_dispatcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'pid' => '',
      'cur' => 'HUF',
      'des-test' => '',
      'des-live' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);
    $form['pid'] = [
      '#type' => 'textfield',
      '#title' => $this->t('CIB Shop ID'),
      '#description' => $this->t('The shop ID received from CIB in email. Has the format "ABC0001".'),
      '#default_value' => $this->configuration['pid'],
      '#required' => TRUE,
    ];
    $form['cur'] = [
      '#type' => 'radios',
      '#title' => $this->t('Currency'),
      '#description' => $this->t('The currency of the payment.'),
      '#options' => [
        'EUR' => 'EUR',
        'HUF' => 'HUF',
      ],
      '#disabled' => TRUE,
      '#default_value' => $this->configuration['cur'],
      '#required' => TRUE,
    ];
    $form['des-test'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Test keyfile path'),
      '#description' => $this->t('The absolute path of the des keyfile used for encryption on the dev site.'),
      '#default_value' => isset($this->configuration['des-test']) ? $this->configuration['des-test'] : '',
      '#required' => TRUE,
    ];
    $form['des-live'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Live keyfile path'),
      '#description' => $this->t('The absolute path of the des keyfile used for encryption on the live site.'),
      '#default_value' => isset($this->configuration['des-live']) ? $this->configuration['des-live'] : '',
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['pid'] = $values['pid'];
      $this->configuration['cur'] = $values['cur'];
      $this->configuration['des-test'] = $values['des-test'];
      $this->configuration['des-live'] = $values['des-live'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $query_pms = $request->query->all();
    // Getting answer from the CIB payment gateway.
    $crypto = 'DATA=' . $query_pms['DATA'] . '&PID=' . $query_pms['PID'] . '&CRYPTO=1';
    $cleartext = $this->encryption->decrypt($crypto);
    parse_str($cleartext, $query21);
    if (($query21['MSGT'] == '21')) {
      // Closing the transaction.
      $amo = (int) $order->getTotalPrice()->getNumber();
      $query32 = [
        'CRYPTO' => 1,
        'MSGT' => 32,
        'PID' => $this->configuration['pid'],
        'TRID' => $query21['TRID'],
        'AMO' => $amo,
      ];

      $url = $this->createUrl($query32);
      $response = $this->httpClient->get($url);
      $vars = [
        '@orderid' => $order->id(),
        '@msgt' => 32,
        '@trid' => $query21['TRID'],
        '@amo' => $amo,
        '@pid' => $this->configuration['pid'],
        '@code' => $response->getStatusCode(),
      ];
      $this->logger->notice($this->t('Request to CIB sent for order @orderid with MSGT @msgt, TRID @trid, AMO @amo and PID @pid. The answer returned a code @code.', $vars));
      // Everything seems to be allright so far.
      if ($response->getStatusCode() == 200) {
        $this->analyseMsgt32Or33($response, $order);
      }
      // Handle timeout.
      elseif ($response->getStatusCode() == 500) {
        $query33 = [
          'CRYPTO' => 1,
          'MSGT' => 33,
          'PID' => $this->configuration['pid'],
          'TRID' => $query32['TRID'],
          'AMO' => $amo,
        ];
        $url = $this->createUrl($query33);
        $response = $this->httpClient->get($url);
        $vars = [
          '@orderid' => $order->order_id,
          '@msgt' => 33,
          '@trid' => $query33['TRID'],
          '@amo' => $amo,
          '@pid' => $this->configuration['pid'],
          '@code' => $response->getStatusCode(),
        ];
        $this->logger->notice($this->t('Request to CIB sent for order @orderid with MSGT @msgt, TRID @trid, AMO @amo and PID @pid. The answer returned a code @code.', $vars));
        if ($response->getStatusCode() == 200) {
          $this->analyseMsgt32Or33($response, $order);
        }
      }
    }
    else {
      $event = new NoCommunication($order);
      $this->eventDispatcher->dispatch(CibEvents::NO_COMMUNICATION, $event);
    }
    return '';
  }

  /**
   * Analyse answer to a 32 and 33 message.
   *
   * @param \Psr\Http\Message\ResponseInterface $response
   *   The response.
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   */
  protected function analyseMsgt32Or33($response, OrderInterface $order) {
    $body = $response->getBody();
    $cleartext2 = $this->encryption->decrypt($response->getBody());
    parse_str($cleartext2, $query);
    $vars = [
      '@orderid' => $order->id(),
      '@trid' => $query['TRID'],
      '@msgt' => $query['MSGT'],
      '@rc' => $query['RC'],
      '@rt' => $query['RT'],
      '@anum' => $query['ANUM'],
    ];
    $this->logger->notice($this->t('Message from CIB for order @orderid. TRID: @trid, MSGT: @msgt, RC: @rc, RT: @rt, ANUM: @anum.', $vars));

    if (empty($query['TRID'])) {
      throw new PaymentGatewayException();
    }
    $payment = $this->entityTypeManager->getStorage('commerce_payment')->loadByRemoteId($query['TRID']);
    if (!empty($payment)) {
      $payment->payment_cib_msg = $query['MSGT'];
      $payment->remote_state = $query['RC'];
      $payment->payment_cib_rt = $query['RT'];
      $payment->payment_cib_anum = $query['ANUM'];
      $end = $payment->payment_cib_end = $this->time->getCurrentTime();
      $start = $payment->payment_cib_start->value;

      if ($query['RC'] == '00') {
        $payment->setState('completed')->save();
      }
      else {
        if ($end - $start > self::TIMEOUT) {
          $payment->setState('pending')->save();
        }
        else {
          $payment->setState('voided')->save();
        }
        $event = new FailedPayment($payment);
        $this->eventDispatcher->dispatch(CibEvents::FAILED_PAYMENT, $event);
        throw new PaymentGatewayException();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createUrl($query, $market_customer = 'market') {
    $cleartext = '';
    foreach ($query as $key => $value) {
      $cleartext .= $key . '=' . rawurlencode($value) . '&';
    }
    $cleartext = substr($cleartext, 0, -1);

    if (!$this->encryption->getKey()) {
      $des = ($this->getMode() == 'test') ? $this->getConfiguration()['des-test'] : $this->getConfiguration()['des-live'];
      $this->encryption->setKeyfile($des);
    }
    $crypto = $this->encryption->encrypt($cleartext);

    if ($market_customer === 'market') {
      if ($this->getMode() === 'live') {
        $base_url = 'http://eki.cib.hu:8090/market.saki';
      }
      else {
        $base_url = 'http://ekit.cib.hu:8090/market.saki';
      }
    }
    else {
      if ($this->getMode() == 'live') {
        $base_url = 'https://eki.cib.hu/customer.saki';
      }
      else {
        $base_url = 'https://ekit.cib.hu/customer.saki';
      }
    }
    return $base_url . '?' . $crypto;
  }

  /**
   * {@inheritdoc}
   */
  public function msgt37Or70(PaymentInterface $payment, $msgt = 37) {
    $config = $payment->getPaymentGateway()->getPluginConfiguration();
    $query = [
      'MSGT' => $msgt,
      'PID' => $config['pid'],
      'TRID' => $payment->getRemoteId(),
      'AMO' => (int) $payment->getAmount()->getNumber(),
    ];
    $url = $this->createUrl($query);
    $response = $this->httpClient->get($url);
    $cleartext = $this->encryption->decrypt($response->getBody()->__toString());
    parse_str($cleartext, $response_array);
    return $response_array;
  }

}
